/***********************************************************************/
/*                                                                     */
/*   svm_struct_api.c                                                  */
/*                                                                     */
/*   Definition of API for attaching implementing SVM learning of      */
/*   structures (e.g. parsing, multi-label classification, HMM)        */ 
/*                                                                     */
/*   Author: Thorsten Joachims                                         */
/*   Date: 03.07.04                                                    */
/*                                                                     */
/*   Copyright (c) 2004  Thorsten Joachims - All rights reserved       */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include <string.h>
#include "svm_struct_common.h"
#include "svm_struct_api.h"

#define MATCH       1
#define SUBST       2
#define DELET       3 
#define INSRT       4 
#define D_OPN       5  /* not part of learning algorithm */
#define D_EXT       6  /* not part of learning algorithm */
#define I_OPN       7  /* not part of learning algorithm */
#define I_EXT       8  /* not part of learning algorithm */
#define GAP         9
#define G_OPN       10 
#define G_EXT       11
#define SUBST_SCALE 12
#define NUM_OPS     12
#define NO_OP      -1
#define NO_OP_BOTH -1
#define NO_OP_A    -2
#define NO_OP_B    -3


/******************************************/
/*   Non-Api Helper functions and types   */
/******************************************/

/* Useful structure for SW algorithm */
typedef struct node{
  double score;
  void  *prev; /* has to be cast to NODE */
  int    operation;
  int    row_index;
  int    column_index;
} NODE;

LABEL smith_waterman(PATTERN x, STRUCTMODEL *sm, 
		     STRUCT_LEARN_PARM *sparm);


int* shrink(int* x, int size){
  /* Copies first 'size' elements of x into a new array. Frees x. */
  int* rVal=(int*)malloc(sizeof(int)*size);
  int  j;
  
  for(j=0;j<size;j++)
    rVal[j]=x[j];
  free(x);
  
  return rVal;
}

int subst_index(int x, int y, STRUCTMODEL *sm){
/* Get index into weight vector for substitution weights. If x or y is
   0 then the index is for character specific insertion/deletion costs. */
  if(x!=0 && y!=0)
    return NUM_OPS +(x-1)*sm->alphabetsize+y;
  else if(x==0 && y!=0)
    return NUM_OPS + sm->alphabetsize*(sm->alphabetsize+1) + y;
  else if(x!=0 && y==0)
    return NUM_OPS + sm->alphabetsize*sm->alphabetsize + x;
  /*else x and y equal 0, this shouldn't ever be the case */
  return 0;
}

int getSubstIndex(int *A,int *B,int *a,int i,STRUCTMODEL *sm)
{
  /* Get substitution index for characters aligned by ith operation in *a. */
  int x=0, y=0, k=0, r=0, s=0;

  for(;k<=i;k++){
    switch(a[k])
      {
      case MATCH: x++; y++; r=A[x-1];s=B[y-1]; break;
      case SUBST: x++; y++; r=A[x-1];s=B[y-1]; break;
      case DELET: x++; r=A[x-1];s=0; break; 
      case INSRT: y++; r=0;s=B[y-1]; break;
      case NO_OP_BOTH: x++; y++; r=0; s=0; break;
      case NO_OP_A: x++; r=0; s=0; break;
      case NO_OP_B: y++; r=0; s=0; break;
      default: break;
      }
  }
  
  return subst_index(r,s,sm);
}

void count_sequences(char *file, long *tot_bundles, long *lseq, long *ll)
{
  /* Finds the total number of examples, length of the longest sequence, and
     the lenght of the longest line in *file */
  
  FILE*  seqfl;
  char*  line;
  long   wol, nol;
  int    counter, nowhere;

  if ((seqfl = fopen (file, "r")) == NULL)
    { perror (file); exit (1); }

  nol_ll(file, &nol,&wol, ll);

  (*ll) += 1;
  *tot_bundles=0;
  *lseq=0;
  line=(char*)malloc(sizeof(char)*(*ll));

  while((!feof(seqfl)) && fgets(line,(int)(*ll),seqfl)) {
    counter =0;
    if(line[0]=='s')
      (*tot_bundles)++;
    
    if(sscanf(line,"%c id:%i seq:%s",(char*)&nowhere,&nowhere,line)<3)
      sscanf(line,"\talign:%s", line);
    
    while(strcmp(line,"0")){
      sscanf(line,"%i.%s",&nowhere,line);
      counter++;
    }

    if(counter > (*lseq)) (*lseq)=counter;
    free(line); line=(char*)malloc(sizeof(char)*(*ll));
  }
  free(line);
}

int sequence_length(int* seq){
  /* finds the length of sequence *seq */
  int i=0;
  while(seq[i++]);
  return i;
}
/**** end helper functions ****/
/******************************/

SAMPLE      read_struct_examples(char *file, STRUCT_LEARN_PARM *sparm){
  /* Reads training examples and returns them in sample. The number of
     examples must be written into sample.n */
  long     tot_bundles, l_seq,ll;
  char    *d_file;
  char    *line;
  FILE    *seqfl;
  int      numBundles=0, *index,i,j;
  PATTERN *tmpBundle;
  SAMPLE   s;
  EXAMPLE *ex;
  int     *validity[3];
  long     max_index=0;

  d_file=file;

  count_sequences(d_file,&tot_bundles,&l_seq,&ll);
  ex=(EXAMPLE*)malloc(sizeof(EXAMPLE)*tot_bundles);

  ll+=2;
  line=(char*)malloc(sizeof(char)*ll);
  /*** initialize validity ***/
  validity[0]=(int*)malloc(sizeof(int)*tot_bundles);
  validity[1]=(int*)malloc(sizeof(int)*tot_bundles);
  validity[2]=(int*)malloc(sizeof(int)*tot_bundles);
  for(i=0;i<3;i++)
    for(j=0;j<tot_bundles;j++)
      validity[i][j]=0;
  /***end***/

  if ((seqfl = fopen (d_file, "r")) == NULL)
    { perror (d_file); exit (1); }

  /**** get maximum index ***/
  while((!feof(seqfl)) && fgets(line,(int)ll,seqfl)) {
    int id;
    char type;
    if(sscanf(line,"%c id:%i seq:%s",&type,&id,line)==3)
      if(id>max_index)
	max_index=id;
    
    free(line);
    line=(char*)malloc(sizeof(char)*ll);
  }
  fclose(seqfl);
  /**** end of max index ***/

  index=(int*)malloc(sizeof(int)*(max_index+1));
  for(i=0;i<max_index+1;i++)
    index[i]=-1;

  if ((seqfl = fopen (d_file, "r")) == NULL)
    { perror (d_file); exit (1); }

  /**** scan sequences ***/
  while((!feof(seqfl)) && fgets(line,(int)ll,seqfl)) 
    {
      int cnt=0,j,id;
      int* seq=malloc(sizeof(int)*(l_seq+2));
      char type;
      int** tmpSeq;

      if(line[0]=='#') /*comment line*/
	continue;
      
      if(sscanf(line,"\talign:%s",line)==1){
	if(type != '+'){
	  printf("\nError in bundle %i: alignment must follow + line\n",id);
	  exit(1);
	}
	type = 'a';
      }
      else if(sscanf(line,"%c id:%i seq:%s",&type,&id,line)==3){
	if(index[id]==-1){
	  index[id]=numBundles++;
	  ex[index[id]].x.id=id;
	  ex[index[id]].y.id=id;
	  ex[index[id]].y.decoy_id=0;
	  ex[index[id]].x.numDecoys=0;
	  ex[index[id]].x.decoy=(int**)malloc(sizeof(int*)*3);
	}
      }
      else{
	printf("\nError in bundle %i: unable to parse line\n",id);
	exit(1);
      }

      
      /***** parse sequence ****/
      while(strcmp(line,"0")){
	sscanf(line,"%i.%s",seq+(cnt++),line);
      }
      *(seq+(cnt))=0;

      /*** fill in appropriate part of bundle ****/
      switch(type)
	{
	case 's': /* native sequence */
	  ex[index[id]].x.original=shrink(seq,cnt+1);
	  if(validity[0][index[id]]++){
	    printf("\nERROR in bundle %i: Multiple original sequences\n",id);
	    exit(1);
	  }
	  break;
	case '+': /* homolog sequence */
	  ex[index[id]].x.homolog=shrink(seq,cnt+1);
	  if(validity[1][index[id]]++)
	    printf("\nERROR in bundle %i: Multiple homologs\n",id);
	  break;
	case '-': /* decoy sequence */
	  tmpBundle=&ex[index[id]].x;
	  tmpBundle->numDecoys++;

	  /* extend array of sequences */
	  tmpSeq=(int**)malloc(sizeof(int*)*(tmpBundle->numDecoys));
	  for(j=0;j<tmpBundle->numDecoys-1;j++)
	    tmpSeq[j]=tmpBundle->decoy[j];
	  free(tmpBundle->decoy);
	  tmpBundle->decoy=(int**)malloc(sizeof(int*)*(tmpBundle->numDecoys));
	  for(j=0;j<tmpBundle->numDecoys-1;j++)
	    tmpBundle->decoy[j]=tmpSeq[j];
	  free(tmpSeq);
	  /* end extend */

	  tmpBundle->decoy[(tmpBundle->numDecoys)-1]=shrink(seq,cnt+1);
	  validity[2][index[id]]++;
	  break;
	case 'a': /* alignment */
	  ex[index[id]].y.align_h=shrink(seq,cnt+1);
	  break;
	default:
	  printf("\nError in bundle %i: invalid type '%c'\n",id,type);
	  exit(1);
	}

      free(line);
      line=malloc(sizeof(char)*ll);

    }
  free(line);
  fclose(seqfl);
  free(index);

  /*** check validity ***/
  for(i=0;i<3;i++)
    for(j=0;j<tot_bundles;j++)
      if(!validity[i][j]){
	printf("\nERROR in bundle %i: incomplete\n",i);
	exit(1);
      }

  free(validity[0]);free(validity[1]);free(validity[2]);
  
  s.examples=ex;
  s.n=tot_bundles;
  return s;
}/* END read_struct_examples */


void        init_struct_model(SAMPLE sample, STRUCTMODEL *sm, 
			      STRUCT_LEARN_PARM *sparm)
{
  EXAMPLE     *b=sample.examples;
  int          n=sample.n;
  int          i,j,k;
  STRUCTMODEL  init_model;

  /* avoid unititalized memory read */
  sm->svm_model=NULL;
  
  /* copy value from command line */
  sm->subst_scale=sparm->subst_scale;
  sm->alphabetsize=sparm->alphabetsize;

  if(!sm->alphabetsize){
    /**** find alphabetsize since it was not initialized ****/
    for(i=0; i<n;i++){
      k=0;
      /* find highest character in original */
      while(b[i].x.original[k] !=0){ 
	if(b[i].x.original[k] > sm->alphabetsize)
	  sm->alphabetsize=b[i].x.original[k];
	k++;
      }
      
      k=0;
      /* find highest character in homolog */
      while(b[i].x.homolog[k] !=0){ 
	if(b[i].x.homolog[k] > sm->alphabetsize)
	  sm->alphabetsize=b[i].x.homolog[k];
	k++;
      }
      
      /*find highest character in decoys */
      for(j=0; j< b[i].x.numDecoys;j++){
	k=0;
	while(b[i].x.decoy[j][k] !=0){
	  if(b[i].x.decoy[j][k] > sm->alphabetsize)
	    sm->alphabetsize=b[i].x.decoy[j][k];
	  k++;
	}
      }
    }
  }
  sm->sizePhi=sm->alphabetsize*(sm->alphabetsize+2) + NUM_OPS;

  if(sparm->fix_subst_matrix){
    /* allocate memory for and initialize fixed substitution matrix */
    sparm->w_subst=(double*)my_malloc(sizeof(double)*(sm->sizePhi+1));
    for(i=0;i<sm->sizePhi+1;i++)
      sparm->w_subst[i]=0;
  }

  if(sparm->initialize_from_file && sparm->fix_subst_matrix){
    /* read model and copy into fixed substitution matrix */
    init_model=read_struct_model(sparm->init_model_file,sparm);

    if(init_model.sizePhi != sm->sizePhi){
      printf("ERROR: Size of phi in initialization file does not match sample's alphabet\n");
      exit(1);
    }

    for(i=0;i<sm->sizePhi+1;i++)
      sparm->w_subst[i]=init_model.w[i];

    free_struct_model(init_model);
  }

  return;
}/* END init_struct_model */

CONSTSET    init_struct_constraints(SAMPLE sample, STRUCTMODEL *sm, 
				    STRUCT_LEARN_PARM *sparm)
{
  /* no initial constraints */
  CONSTSET c;
  c.lhs=NULL;
  c.rhs=NULL;
  c.m=0;
  return c;
}

LABEL       classify_struct_example(PATTERN x, STRUCTMODEL *sm, 
				    STRUCT_LEARN_PARM *sparm)
{
    /**************************************************************************/
   /* Find the optimal alignment for the original and homolog sequence in x. */
  /**************************************************************************/
  return smith_waterman(x,sm,sparm);
}

LABEL       find_most_violated_constraint_slackrescaling(PATTERN x, LABEL y, 
						     STRUCTMODEL *sm, 
						     STRUCT_LEARN_PARM *sparm)
{
     /******************************************************************/
    /* Find x.decoy[i] with highest scoring alignment with x.original,*/
   /* given parameters in sm->w. Returned as LABEL struct.           */
  /******************************************************************/

  int      i=0;
  LABEL    temp_label;
  double   score_y=0;
  double   temp_score;
  PATTERN  x_d;
  SVECTOR* temp_phi;
  LABEL    ybar;

  ybar.align_h=NULL;

  for(;i<x.numDecoys;i++)
    {
      x_d.original=x.original;
      x_d.homolog=x.decoy[i];

      temp_label=classify_struct_example(x_d, sm, sparm);
      temp_label.decoy_id=0;
      temp_phi=phi(x_d,temp_label,sm,sparm);
      temp_score=sprod_ns(sm->w,temp_phi);
      free_svector(temp_phi);

      if(score_y<=temp_score){
	score_y=temp_score;
	free_label(ybar);
	ybar=temp_label;
	ybar.decoy_id=i+1;
      }
      else
	free_label(temp_label);
    }

  return ybar;
}

LABEL       find_most_violated_constraint_marginrescaling(PATTERN x, LABEL y, 
						     STRUCTMODEL *sm, 
						     STRUCT_LEARN_PARM *sparm)
{
  /* Since loss is always zero, this is the same as slackrescaling */
  return find_most_violated_constraint_slackrescaling(x,y,sm,sparm);
}

int         empty_label(LABEL y)
{
   return !y.align_h;
}

SVECTOR     *phi(PATTERN x, LABEL y, STRUCTMODEL *sm,
		 STRUCT_LEARN_PARM *sparm)
{

      /***********************************************************/
     /* Creates a vector with each value corresponding to the   */
    /* number of times a particular operation is performed.    */
   /* Returned as SVECTOR struct (see svm_common.h)           */
  /***********************************************************/

  int      *A,*B,*a;
  WORD     *features;
  int      i,j,length_alignment=0;
  int      sizePhi=sm->sizePhi;
  SVECTOR  *fvec;  

  features=(WORD*)my_malloc(sizeof(WORD)*(sizePhi+1));
  A=x.original;
  B=x.homolog;
  if(y.decoy_id != 0)
    B=x.decoy[y.decoy_id-1];
  
  a=y.align_h;

  /* initialize everything to 0 */
  for(i=0; i<sizePhi;i++){
    features[i].wnum=i+1; 
    features[i].weight=0;
  }
  features[sizePhi].wnum=0;
  features[sizePhi].weight=0;

    /**********************************************************/
   /* Count the occurrence of each operation in the alignment*/
  /**********************************************************/
  for(i=0; a[i] !=0;i++){
    switch(a[i])
      {
      case INSRT:
      case DELET:
	if(sparm->gap_extension || sparm->use_all_param){     /* if gap extension turned on or */
	  if(i>0 && a[i-1]==a[i])                            /* use_all_param then include in */
	    features[G_EXT-1].weight++;                     /* the count G_OPN & G_EXT.      */
	  else
	    features[G_OPN-1].weight++;
	}
	
	if((sparm->ind_gap_costs < 2 && !sparm->gap_extension) || sparm->use_all_param)
	  features[GAP-1].weight++;
	
	if(sparm->ins_del_score  || sparm->use_all_param)
	  features[a[i]-1].weight++;

	if(sparm->ind_gap_costs > 0  || sparm->use_all_param)   /* include ind gap costs in phi */
	  if((j=getSubstIndex(A,B,a,i,sm)))                    /* if the ind_gap flag is set   */
	    features[j-1].weight += sm->subst_scale;             

	length_alignment++;
	break;
      case SUBST:
      case MATCH:
	features[a[i]-1].weight++;
	if(sparm->fix_subst_matrix)
	  /* there is still a bug in the fix_subst_matrix option */
	  features[SUBST_SCALE-1].weight += sparm->w_subst[getSubstIndex(A,B,a,i,sm)];
	else{                                               /* if subst matrix is to be fixed  */
	  if((j=getSubstIndex(A,B,a,i,sm)))                /* then leave the corresponding    */
	    features[j-1].weight += sm->subst_scale;      /* weights in phi at 0.            */
	}

	length_alignment++;
	break;
      }
  }

  fvec=create_svector(features,"NULL",1.0);
  free(features);
  return(fvec);
}

double      loss(LABEL y, LABEL ybar, STRUCT_LEARN_PARM *sparm)
{
  /* loss for correct label y and predicted label ybar. The loss for
     y==ybar has to be zero. sparm->loss_function is set with the -l option. */
  if(sparm->loss_function == 0) { /* type 0 loss: 0/1 loss */
    return 1;                     /* return 0, if y==ybar. return 1 else */
  }
  else {
    /* Put your code for different loss functions here. But then
       find_most_violated_constraint_???(x, y, sm) has to return the
       highest scoring label with the largest loss. */
  }
  return 1;
}

void        print_struct_learning_stats(SAMPLE sample, STRUCTMODEL *sm,
					CONSTSET cset, double *alpha, 
					STRUCT_LEARN_PARM *sparm)
{
  /* This function is called after training and allows final touches to
     the model sm. But primarly it allows computing and printing any
     kind of statistic (e.g. training error) you might want. */
}

void        write_struct_model(char *file, STRUCTMODEL *sm, 
			       STRUCT_LEARN_PARM *sparm)
{
  /* Writes structural model sm to file file. */
  FILE   *m_file;
  int    i,j;
  double k;

  if ((m_file = fopen (file, "w")) == NULL)
  { perror (file); exit (1); }

  fprintf(m_file, "ALPHABETSIZE: %i\n",sm->alphabetsize);
  if(sparm->fix_subst_matrix)
    sm->subst_scale=sm->w[SUBST_SCALE];
  if(sm->subst_scale != 1)
    fprintf(m_file, "SUBST SCALE: %f\n", sm->subst_scale);

  sm->w[MATCH] ? fprintf(m_file, "MATCH: %f\n",        sm->w[MATCH]) : 0;
  sm->w[SUBST] ? fprintf(m_file, "SUBST: %f\n",        sm->w[SUBST]) : 0;
  sm->w[GAP]   ? fprintf(m_file, "GAP: %f\n",          sm->w[GAP]  ) : 0;
  sm->w[G_OPN] ? fprintf(m_file, "GAP OPEN: %f\n",     sm->w[G_OPN]) : 0;
  sm->w[G_EXT] ? fprintf(m_file, "GAP EXT: %f\n",      sm->w[G_EXT]) : 0;
  sm->w[INSRT] ? fprintf(m_file, "INSERT: %f\n",       sm->w[INSRT]) : 0;
  sm->w[I_OPN] ? fprintf(m_file, "INSERT OPEN: %f\n",  sm->w[I_OPN]) : 0;
  sm->w[I_EXT] ? fprintf(m_file, "INSERT EXT: %f\n",   sm->w[I_EXT]) : 0;
  sm->w[DELET] ? fprintf(m_file, "DELETE: %f\n",       sm->w[DELET]) : 0;
  sm->w[D_OPN] ? fprintf(m_file, "DELETE OPEN: %f\n",  sm->w[D_OPN]) : 0;
  sm->w[D_EXT] ? fprintf(m_file, "DELETE EXT: %f\n",   sm->w[D_EXT]) : 0;
  
  for(i=1;i<=sm->alphabetsize;i++)
    for(j=1;j<=sm->alphabetsize;j++){
      if(sparm->fix_subst_matrix)
	k=sparm->w_subst[NUM_OPS+(i-1)*sm->alphabetsize+j];
      else
	k=sm->w[NUM_OPS+(i-1)*sm->alphabetsize+j];

      if(i==j)
	k ? fprintf(m_file,"M(%i): %f\n",i,(float)k):0;
      else
	k ? fprintf(m_file,"S(%i,%i): %f\n",i,j,(float)k):0;
    }

  for(i=1;i<=sm->alphabetsize;i++){/*print individual gap costs */
    k=sm->w[subst_index(i,0,sm)];
    k ? fprintf(m_file,"D(%i): %f\n",i,(float)k):0;
  }
  for(i=1;i<=sm->alphabetsize;i++){
    k=sm->w[subst_index(0,i,sm)];
    k ? fprintf(m_file,"I(%i): %f\n",i,(float)k):0;
  }
    
  /* HACK: Freeing STRUCT_LEARN_PARM is not built into the API so to
     avoid a memory leak, I'm going to free any allocated memory here. */
  if(sparm->fix_subst_matrix)
    if(sparm->w_subst) free(sparm->w_subst);
  
  fclose(m_file);
}

STRUCTMODEL read_struct_model(char *file, STRUCT_LEARN_PARM *sparm)
{
  /* Reads structural model sm from file.  Returned STRUCTMODEL is filled in 
     according to contents of file and is not checked against sparm. For instance
     the returned model may include gap extension values but sparm might have
     gap extension flagged OFF. A number of parameters such as this are read into
     STRUCT_LEARN_PARM from the command line and then copied to STRUCT_MODEL. */

  STRUCTMODEL sm;
  FILE       *m_file;
  long        nol, wol, ll;
  char       *line;
  int         i,j;
  
  nol_ll(file, &nol,&wol, &ll);

  ll =ll +3;
  if ((m_file = fopen (file, "r")) == NULL)
    { perror (file); exit (1); }

  sm.svm_model=NULL;

  /**** read sizePhi - for backwards compatability****/
  line=(char*)malloc(sizeof(char)*ll);

  if((!feof(m_file)) && fgets(line,(int)ll,m_file))
    if(sscanf(line, "SIZEPHI: %i",(int*)&sm.sizePhi)==1){
      if(struct_verbosity>1)
	printf("\nWARNING: Old model file. SIZEPHI no longer part of format.\n");
      !feof(m_file) && fgets(line,(int)ll,m_file);
    }
 
  /***** read alphabetsize ****/
  if(sscanf(line, "ALPHABETSIZE: %i",&sm.alphabetsize)!=1)
    {printf("ERROR Reading ModelFile: First line must contain ALPHABETSIZE\n"); exit(1);}
  
  sm.sizePhi=NUM_OPS+sm.alphabetsize*(sm.alphabetsize+2);/*includes space for individual gap costs*/
  sm.subst_scale=1;

  /**** read weight vector ****/
  sm.w=(double*)malloc(sizeof(double)*(sm.sizePhi+1));
  for(i=0;i<sm.sizePhi+1;i++)
    sm.w[i]=0;

  while((!feof(m_file)) && fgets(line,(int)ll,m_file))
    {
      if(sscanf(line, "SUBST SCALE: %f",&sm.subst_scale) == 1){
	if(struct_verbosity > 2)
	  printf("Substitution matrix scale set to %f\n", sm.subst_scale);
      }      
      /*********** READ OTHER PARAMETERS ***********/
      else if(sscanf(line, "%i:%s",&i, line)==2) /* old format -- not valid anymore */
	{printf("Old model format! Exiting....\n"); exit(1);}
      
      /* match and substitution parameters */
      else if(sscanf(line, "M(%i):%s",&i,line)==2)
	sm.w[subst_index(i,i,&sm)]=atof(line);
      else if(sscanf(line, "MATCH:%s",line)==1)
	sm.w[MATCH]=atof(line);
      else if(sscanf(line, "S(%i,%i):%s",&i,&j,line)==3)
	sm.w[subst_index(i,j,&sm)]=atof(line);
      else if(sscanf(line, "SUBST:%s",line)==1)
	sm.w[SUBST]=atof(line);
      
      /* insertion parameters */
      else if(sscanf(line, "I(%i):%s",&i,line)==2)
	sm.w[subst_index(0,i,&sm)]=atof(line);
      else if(sscanf(line, "INSERT:%s",line)==1)
	sm.w[INSRT]=atof(line);
      else if(sscanf(line, "INSERT OPEN:%s",line)==1)
	sm.w[I_OPN]=atof(line);
      else if(sscanf(line, "INSERT EXT:%s",line)==1)
	sm.w[I_EXT]=atof(line);
      
      /* deletion parameters */
      else if(sscanf(line, "D(%i):%s",&i,line)==2)
	sm.w[subst_index(i,0,&sm)]=atof(line);
      else if(sscanf(line, "DELETE:%s",line)==1)
	sm.w[DELET]=atof(line);
      else if(sscanf(line, "DELETE OPEN:%s",line)==1)
	sm.w[D_OPN]=atof(line);
      else if(sscanf(line, "DELETE EXT:%s",line)==1)
	sm.w[D_EXT]=atof(line);
      
      /* general gap parameters */
      else if(sscanf(line, "GAP:%s",line)==1)
	sm.w[GAP]=atof(line);
      else if(sscanf(line, "GAP OPEN:%s",line)==1)
	sm.w[G_OPN]=atof(line);
      else if(sscanf(line, "GAP EXT:%s",line)==1)
	sm.w[G_EXT]=atof(line);
      else
	{printf("ERROR reading weight vector in model file\n"); exit(1);}
    }
  

  /**** done reading ****/
  fclose(m_file);
  free(line);
  sm.subst_scale=1;
  sm.svm_model=NULL;
  return sm;
}

void        write_label(FILE *fp, LABEL y)
{
  /* Writes label y to file handle fp. */
  while(*y.align_h)
    fprintf(fp,"%i.",*(y.align_h++));
  fprintf(fp,"0\n");
} 

void        free_pattern(PATTERN x) {
  /* Free's the memory of x. */
  int i;

  free(x.original);
  free(x.homolog);
  for(i=0;i<x.numDecoys;i++)
    free(x.decoy[i]);
  free(x.decoy);
}

void        free_label(LABEL y) {
  /* Frees the memory of y. */
  if(y.align_h) free(y.align_h);
}

void        free_struct_model(STRUCTMODEL sm) 
{
  /* Frees the memory of model. */
  if(sm.svm_model)                        /* in this case sm.w should point */
    free_model(sm.svm_model,1);          /* to sm.svm_model.lin_weights.   */
  else                                  /* else                           */
    if(sm.w) free(sm.w);               /* need to free sm.w separately   */
}

void        free_struct_sample(SAMPLE s)
{
  /* Frees the memory of sample s. */
  int i;
  for(i=0;i<s.n;i++) { 
    free_pattern(s.examples[i].x);
    free_label(s.examples[i].y);
  }
  free(s.examples);
}

void        print_struct_help()
{
  /* Prints a help text that is appended to the common help text of
     svm_struct_learn. */
  printf("         --s  float -> Weight of substitution matrix (default 1)\n");
  printf("         --a  int   -> Manually specify the size of the alphabet.  Do this\n");
  printf("                       if not all characters appear in training set. A value of 0\n");
  printf("                       sets the alphabet size to the greates value appearing in the\n");
  printf("                       training set. (default 0)\n");
  printf("         --i [0-2]  -> Individual gap costs\n");
  printf("                              0: Only use general, non-character-specific gap cost (default)\n");
  printf("                              1: Use individual gap costs in addition to general gap cost\n");
  printf("                              2: Only use individual gap costs.\n");
  printf("                       each character(default 0)\n");
  printf("         --g [0,1]  -> Gap extension (default 0)\n");
  printf("         --h [0,1]  -> Use separate score for insert and delete (default 0)\n");

  /* fixing the substitution matrix doesn't work quite right, so i'm hiding it from the user for now*/
  /*printf("         --f [0,1]  -> Fix substitution matrix to initial weights, i.e. \n");
    printf("                       Optimization will not include these variables. (default 0)\n");*/
}

void         parse_struct_parameters(STRUCT_LEARN_PARM *sparm)
{
  /* Parses the command line parameters that start with -- */
  int i;

  sparm->alphabetsize=0;
  sparm->subst_scale=1;
  sparm->ind_gap_costs=0;
  sparm->fix_subst_matrix=0;
  sparm->gap_extension=0;
  sparm->ins_del_score=0;
  sparm->use_all_param=0;

  for(i=0;(i<sparm->custom_argc) && ((sparm->custom_argv[i])[0] == '-');i++) {
    switch ((sparm->custom_argv[i])[2]) 
      { 
      case 'a': i++; sparm->alphabetsize=atoi(sparm->custom_argv[i]); break;
      case 's': i++; sparm->subst_scale=atof(sparm->custom_argv[i]); break;
      case 'i': i++; sparm->ind_gap_costs=atoi(sparm->custom_argv[i]); break;
      case 'g': i++; sparm->gap_extension=atoi(sparm->custom_argv[i]); break;
      case 'f': i++; sparm->fix_subst_matrix=atoi(sparm->custom_argv[i]); break; /*doesn't work*/
      case 'h': i++; sparm->ins_del_score=atoi(sparm->custom_argv[i]); break;
      }
  }

  /*** error checking ***/
  if(sparm->fix_subst_matrix != 0 && sparm->fix_subst_matrix != 1)
    {printf("Valid values for --f are 0(off) or 1(on)\n");exit(1);}
  if(sparm->gap_extension != 0 && sparm->gap_extension != 1)
    {printf("Valid values for --g are 0(off) or 1(on)\n");exit(1);}
  if(sparm->ind_gap_costs < 0 || sparm->ind_gap_costs > 2)
    {printf("Valid values for --i are 0,1,2\n");exit(1);}
  if(sparm->ins_del_score != 0 && sparm->ins_del_score != 1)
    {printf("Valid values for --h are 0(off) or 1(on)\n"); exit(1);}
  if(sparm->alphabetsize < 0)
    {printf("Value for --a must be non-negative\n");exit(1);}

  if(sparm->fix_subst_matrix ==1 && sparm->subst_scale != 1)
    if(struct_verbosity >0)
      printf("Setting substitution matrix weight to 1.\n");
  if(sparm->fix_subst_matrix ==1 && sparm->initialize_from_file != 1)
    if(struct_verbosity >0)
      printf("WARNING: No initialization file given. Substitution matrix will be fixed at 0!\n");

  /** if someone actually uses --f then spit out a warning */
  if(sparm->fix_subst_matrix ==1){
    printf("Somehow you've decided to use the --f option. This option doesn't\n");
    printf("work quite right so there's no telling what kind of results you'll\n");
    printf("get. Good Luck\n");
  }
}

/***** End API methods ****/




  /************************************************************/
 /***************** SMITH WATERMAN ALGORITHM *****************/
/************************************************************/

LABEL smith_waterman(PATTERN x, STRUCTMODEL *sm, 
		      STRUCT_LEARN_PARM *sparm){
      /***************************************************************/
     /* Finds the highest scoring alignment between x.original and  */
    /* x.homolog, given the parameter values in sm->w. Returned in */
   /* LABEL struct. Uses gap extension                            */
  /***************************************************************/

  int    *A, *B, *a;
  int     i,j,maxA=0,maxB=0, minA=0, minB=0;
  double *w, subst_ij, temp;
  NODE  **S, **I, **D;
  NODE   *endSequence, *temp_node;
  LABEL   y;
  int     lengthA=0, lengthB=0;

  /*** for debugging 
  double temp_score=0;
  int match_count=0;
  int sub_count=0;
  int gap_count=0;
  ******************/

  w=sm->w;
  A=x.original;
  B=x.homolog;

  a=malloc(sizeof(int)*(sequence_length(A) + sequence_length(B)+1));

  /* get lengths of each sequence */
  for (; A[lengthA] != 0; lengthA++); 
  for (; B[lengthB] != 0; lengthB++); 

  /* initialize matrix */
  S=(NODE**)malloc(sizeof(NODE*)*(lengthA+1));
  I=(NODE**)malloc(sizeof(NODE*)*(lengthA+1));
  D=(NODE**)malloc(sizeof(NODE*)*(lengthA+1));
  for(i=0;i<lengthA+1; i++){
    S[i]=(NODE*)malloc(sizeof(NODE)*(lengthB+1));
    I[i]=(NODE*)malloc(sizeof(NODE)*(lengthB+1));
    D[i]=(NODE*)malloc(sizeof(NODE)*(lengthB+1));
  }
    

  for(i=0; i<lengthA+1; i++){
    S[i][0].score=0;             D[i][0].score=0;          I[i][0].score=0;
    S[i][0].prev=NULL;           D[i][0].prev=NULL;        I[i][0].prev=NULL;
    S[i][0].row_index=0;         D[i][0].row_index=0;      I[i][0].row_index=0;
    S[i][0].column_index=0;      D[i][0].column_index=0;   I[i][0].column_index=0;
  }
  for(i=0;i<lengthB+1;i++){
    S[0][i].score=0;             D[0][i].score=0;          I[0][i].score=0;
    S[0][i].prev=NULL;           D[0][i].prev=NULL;        I[0][i].prev=NULL;
    S[0][i].row_index=0;         D[0][i].row_index=0;      I[0][i].row_index=0;
    S[0][i].column_index=0;      D[0][i].column_index=0;   I[0][i].column_index=0;
  }

  endSequence=&S[0][0];


  /*fill in matrix */
  for(i=1; i<lengthA+1; i++)
    for(j=1; j<lengthB+1; j++)
      {
	S[i][j].row_index=D[i][j].row_index=I[i][j].row_index=i;
	S[i][j].column_index=D[i][j].column_index=I[i][j].column_index=j;
	S[i][j].score=D[i][j].score=I[i][j].score=0;
	S[i][j].prev=D[i][j].prev=I[i][j].prev=NULL;
	
	/***** DELETION MATRIX *******/
	D[i][j].operation=DELET;
	if((temp=D[i-1][j].score + w[G_EXT] + w[GAP] + w[DELET] + w[D_EXT] +
	  w[subst_index(A[i-1],0,sm)]) > D[i][j].score){
	  D[i][j].score=temp;
	  D[i][j].prev=&D[i-1][j];
	}
	if((temp=S[i-1][j].score + w[G_OPN] + w[GAP] + w[DELET] + 
	   w[D_OPN] + w[subst_index(A[i-1],0,sm)])> D[i][j].score){
	  D[i][j].score=temp;
	  D[i][j].prev=&S[i-1][j];
	}
	if((temp=I[i-1][j].score + w[G_OPN] + w[GAP] + w[DELET] + w[D_OPN] + 
	   w[subst_index(A[i-1],0,sm)]) > D[i][j].score){
	  D[i][j].score=temp;
	  D[i][j].prev=&I[i-1][j];
	}

	/***** INSERTION MATRIX ******/
	I[i][j].operation=INSRT;
	if((temp=I[i][j-1].score + w[G_EXT] + w[GAP] + w[INSRT] + w[I_EXT] + 
	   w[subst_index(0,B[j-1],sm)]) > I[i][j].score){
	  I[i][j].score=temp;
	  I[i][j].prev=&I[i][j-1];
	}
	if((temp=S[i][j-1].score + w[G_OPN] + w[GAP]  + w[INSRT] + w[I_OPN] + 
	   w[subst_index(0,B[j-1],sm)]) > I[i][j].score){
	  I[i][j].score=temp;
	  I[i][j].prev=&S[i][j-1];
	}
	if((temp=D[i][j-1].score + w[G_OPN] + w[GAP]  + w[INSRT] + w[I_OPN] + 
	   w[subst_index(0,B[j-1],sm)])> I[i][j].score){
	  I[i][j].score=temp;
	  I[i][j].prev=&D[i][j-1];
	}

	/***** SUBSTITUTION MATRIX ***/
	S[i][j].operation= (A[i-1]==B[j-1] ? MATCH : SUBST);
	if(sparm->fix_subst_matrix)
	  subst_ij=sm->w[SUBST_SCALE]*sparm->w_subst[subst_index(A[i-1],B[j-1],sm)]+w[S[i][j].operation];
	else
	  subst_ij=sm->subst_scale*w[subst_index(A[i-1],B[j-1],sm)]+w[S[i][j].operation];

	if((temp=S[i-1][j-1].score+subst_ij) > S[i][j].score){
	  S[i][j].score=temp;
	  S[i][j].prev=&S[i-1][j-1];
	}
	if((temp=D[i-1][j-1].score+subst_ij) > S[i][j].score){
	  S[i][j].score=temp;
	  S[i][j].prev=&D[i-1][j-1];
	}
	if((temp=I[i-1][j-1].score+subst_ij) > S[i][j].score){
	  S[i][j].score=temp;
	  S[i][j].prev=&I[i-1][j-1];
	}

	/* record information if new maximum */
	if(I[i][j].score > endSequence->score){
	  endSequence=&I[i][j];
	  maxA=i; maxB=j;
	}
	if(D[i][j].score > endSequence->score){
	  endSequence=&D[i][j];
	  maxA=i; maxB=j;
	}
	if(S[i][j].score > endSequence->score){
	  endSequence=&S[i][j];
	  maxA=i; maxB=j;
	}
      }

  /*************************************/
  /* Retrieve alignment (very messy)   */
  /*************************************/
  /*get length of alignment*/
  j=0;
  temp_node=endSequence;
  while(endSequence->prev){
    minA=endSequence->row_index;
    minB=endSequence->column_index;
    j++; endSequence=(NODE*)endSequence->prev;
  }

  endSequence=temp_node;

  /*fill in beginning NO_OPs */
  i=j;
  if(minA<minB){
    for(j=0;j<(minB-minA);j++)
      a[j]=NO_OP_B;
    for(;j<minB;j++)
      a[j]=NO_OP_BOTH;
  }
  else{
    for(j=0;j<(minA-minB);j++)
      a[j]=NO_OP_A;
    for(;j<minA;j++)
      a[j]=NO_OP_BOTH;
  }

  i=i+j-1;
  j=i;

  if(j<0) /* if empty alignment */
    j=0;

  /* fill in local alignment */
  while(endSequence->prev){
    a[--i]= endSequence->operation;

    /**** For debugging fix_subst_matrix ****
    if(endSequence->operation == MATCH || endSequence->operation == SUBST){
      if(sparm->fix_subst_matrix)
	temp_score += sm->w[SUBST_SCALE]*sparm->w_subst[subst_index(A[endSequence->row_index-1],B[endSequence->column_index-1],sm)];
      else
	temp_score += sm->subst_scale*sm->w[subst_index(A[endSequence->row_index-1],B[endSequence->column_index-1],sm)];

      if(endSequence->operation ==MATCH)
	match_count++;
      if(endSequence->operation == SUBST)
	sub_count++;

    }
    if(endSequence->operation == INSRT || endSequence->operation == DELET)
    gap_count++;
    ******** End Debugging stuff ***********/
    endSequence=(NODE*)endSequence->prev;
  }

  a[j]=0; /*terminator*/
  
  /*free DP matrices */
  for(i=0;i<lengthA+1;i++){
    free(S[i]);
    free(D[i]);
    free(I[i]);
  }
  free(S);
  free(D);
  free(I);

  y.align_h=a;
  y.decoy_id=0;
  return y;
} /**** END Smith Waterman ****/
