/***********************************************************************/
/*                                                                     */
/*   svm_align_classify.c                                              */
/*                                                                     */
/*   Classification module of Support Vector Machine for Alignment     */
/*                                                                     */
/*   Author: Philip Zigoris                                            */
/*   Date: 18.01.04                                                    */
/*                                                                     */
/*   Copyright (c) 2003  Thorsten Joachims - All rights reserved       */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/************************************************************************/

# include "svm-light/svm_common.h"
# include "svm_struct_common.h"
# include "svm_struct_api.h"
# include "svm_struct_api_types.h"

char testfile[200];
char modelfile[200];
char predictionsfile[200];

void read_input_parameters(int, char **, char *, char *, char *, long *, 
			   long *);
void print_help(void);

int main (int argc, char* argv[])
{

  /* init */
  SAMPLE            s;
  STRUCT_LEARN_PARM sparm;
  STRUCTMODEL       sm;
  LABEL             y,ybar, temp_label;
  SVECTOR          *fy, *fybar, *temp_phi;
  EXAMPLE          *ex;
  FILE             *predfl;
  PATTERN           x_d;
  
  long    correct=0,incorrect=0,no_accuracy=0, pred_format;
  int     i,j;
  double  delta_error=0, t1, runtime=0, doc_label, temp_score;
  double  score_y,score_ybar, *tau_error, sumTau=0, sumTauSq=0;
  
  
  read_input_parameters(argc,argv,testfile,modelfile,predictionsfile,
			&verbosity,&pred_format);

  /****** Load test sequences ******/
  if(verbosity>=2) {
    printf("Scanning examples..."); fflush(stdout);
  } 
  s=read_struct_examples(testfile,&sparm);
  ex=s.examples;
  if(verbosity>=2) {
    printf("done\n"); fflush(stdout);
  }
  /******* end of sequence read *****/

  /******* Load Model *******/
  if(verbosity>=2) {
    printf("Loading Model..."); fflush(stdout);
  } 
  sm=read_struct_model(modelfile, &sparm);
  if(verbosity>=2) {
    printf("done\n"); fflush(stdout);
  }
  /****** end of model read ********/

  /** open output file **/
  if ((predfl = fopen (predictionsfile, "w")) == NULL)
  { perror (predictionsfile); exit (1); }
  /**** end open output***/

  /* Here we set sparm to use all available parameters.*/
  sparm.fix_subst_matrix=0; 
  sparm.use_all_param=1;
  
  if(verbosity>=2) {
    printf("Classifying test examples.."); fflush(stdout);
  }
  
  correct=incorrect=0;
  tau_error=(double*)my_malloc(sizeof(double)*s.n);

  /** Test Performance **/
  for(i=0;i<s.n;i++)
    {
      if(verbosity>=2) /* print update of status */
	if(i % 10 == 0)
	  printf("%i..",i); fflush(stdout);
      
      t1=get_runtime();
      
      /*test for correctness*/
      y=classify_struct_example(ex[i].x, &sm, &sparm);
      fy=phi(ex[i].x,y,&sm,&sparm);
      score_y=sprod_ns(sm.w,fy);
      
      score_ybar=0;
      tau_error[i]=0;
      ybar.align_h=NULL;/*to avoid memory error*/
      for(j=0;j<ex[i].x.numDecoys;j++){
	x_d.original=ex[i].x.original;
	x_d.homolog=ex[i].x.decoy[j];
	
	temp_label=classify_struct_example(x_d, &sm, &sparm);
	temp_label.decoy_id=j+1;
	temp_phi=phi(ex[i].x,temp_label,&sm,&sparm);
	temp_score=sprod_ns(sm.w,temp_phi);
	free_svector(temp_phi);
	
	if(score_y<=temp_score) tau_error[i]++;
	if(score_ybar<=temp_score)
	  {
	    score_ybar=temp_score;
	    free_label(ybar);
	    ybar=temp_label;
	    ybar.decoy_id=j+1;
	  }
	else
	  free_label(temp_label);
      }
      
      if(empty_label(ybar)) {
	if(struct_verbosity>=2)
	  printf("no-incorrect-found(%i) ",(int)i);
	continue;
      }
      
      /**** get phi(y)-phi(ybar) ****/
      fybar=phi(ex[i].x,ybar,&sm,&sparm);
      doc_label=score_y-score_ybar;
      
      tau_error[i]=tau_error[i]/ex[i].x.numDecoys*100;
      
      runtime+=(get_runtime()-t1);
      
      if(doc_label<=0) incorrect++;	else correct++;
      
      /*** Print info out to file ***/
      if(pred_format==1){ /* print the value of decision function */
	fprintf(predfl,"%.8g\n",doc_label);
      }
      else if(pred_format==0){ /*print patterns to file*/
	write_label(predfl,y);
      }

      free_svector(fy);
      free_svector(fybar);
      free_label(y);
      free_label(ybar);
    }

  delta_error=incorrect/(double)s.n;
  /*** end for loop ***/

  /**** print performance info ****/
  if(verbosity>=2) {
    printf("done\n");
    printf("Runtime (without IO) in cpu-seconds: %.2f\n",
	   (float)(runtime/100.0)); /* see note in svm_classify.c */
  }

  /****** print performance information *****/
  if((!no_accuracy) && (verbosity>=1)) {
    for(i=0;i<s.n;i++){
      if(verbosity>=3)
	printf("\tTau[%i]: %.4f%%\n",i+1, tau_error[i]);
      sumTau += tau_error[i];
      sumTauSq += tau_error[i]*tau_error[i];
    }


    /*printf("Tau Error: %.4f%%(variance %.4f)\n",sumTau/s.n,
      ((s.n*sumTauSq)-(sumTau*sumTau))/(s.n*(s.n-1)));*/
    printf("Tau Error: %.4f%%\n",sumTau/s.n);
    printf("Delta Error: %.4f%%\n",(float)(delta_error*100));
  }
  /***** end print *****/
  free(tau_error);
  free_struct_sample(s);
  free_struct_model(sm);
  return(0);
}
/******end main ****/

void read_input_parameters(int argc, char **argv, char *docfile, 
			   char *modelfile, char *predictionsfile, 
			   long int *verbosity, long int *pred_format)
{
  long i;
  
  /* set default */
  strcpy (modelfile, "svm_align_model");
  strcpy (predictionsfile, "svm_align_margins.out"); 
  (*verbosity)=2;
  (*pred_format)=0; /* default pred_format */

  for(i=1;(i<argc) && ((argv[i])[0] == '-');i++) {
    switch ((argv[i])[1]) 
      { 
      case 'h': print_help(); exit(0);
      case 'v': i++; (*verbosity)=atol(argv[i]); break;
      case 'f': i++; (*pred_format)=atol(argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",argv[i]);
	       print_help();
	       exit(0);
      }
  }
  if((i+1)>=argc) {
    printf("\nNot enough input parameters!\n\n");
    print_help();
    exit(0);
  }
  strcpy (docfile, argv[i]);
  strcpy (modelfile, argv[i+1]);
  if((i+2)<argc) {
    strcpy (predictionsfile, argv[i+2]);
  }
  if(((*pred_format) != 0) && ((*pred_format) != 1)) {
    printf("\nOutput format can only take the values 0 or 1!\n\n");
    print_help();
    exit(0);
  }
}

void print_help(void)
{
  printf("\nSVM-align %s: Support Vector Machine for Alignment, Classification Program     %s\n",VERSION,VERSION_DATE);
  copyright_notice();
  printf("   usage: svm_align_classify [options] example_file model_file output_file\n\n");
  printf("options: -h         -> this help\n");
  printf("         -v [0..3]  -> verbosity level (default 2)\n");
  printf("         -f [0,1]   -> 0: output alignments(default)\n");
  printf("                    -> 1: output the margin of the alignment\n");
}




